package be.iminds.ilabt.jfed.experimenter_gui.timeline.view;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.util.ExperimentUtils;
import be.iminds.ilabt.jfed.experimenter_gui.slice.DetailsTabHost;
import be.iminds.ilabt.jfed.experimenter_gui.slice.NodePropertiesDialogFactory;
import be.iminds.ilabt.jfed.experimenter_gui.timeline.controller.TimelineController;
import be.iminds.ilabt.jfed.experimenter_gui.timeline.model.TimelineModel;
import be.iminds.ilabt.jfed.experimenter_gui.timeline.model.TimelineModelFactory;
import be.iminds.ilabt.jfed.experimenter_gui.timeline.view.output.ControllerOutput;
import be.iminds.ilabt.jfed.experimenter_gui.util.ImageUtil;
import be.iminds.ilabt.jfed.gui_preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec_fx.model.controller.ExperimentBarrierSegment;
import be.iminds.ilabt.jfed.rspec_fx.model.controller.ExperimentCommand;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXRspecNode;
import javafx.beans.value.ObservableBooleanValue;
import javafx.geometry.Orientation;
import javafx.scene.control.SplitPane;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.Objects;

/**
 * @author Cedric
 */
public class TimelineView extends SplitPane implements TimelineComponent {
    private static final Logger LOG = LoggerFactory.getLogger(TimelineView.class);

    @Nonnull private final FXModelRspec modelRspec;
    @Nonnull private final TimelineModel model;
    @Nonnull private final TimelineController controller;
    @Nonnull private final TimelineTimeline experimentController;
    @Nonnull private final TimelineTreeView experimentList;
    @Nullable private final ControllerOutput experimentOutput;

    /**
     * Constructor which initiates ("offline") editor-view of timeline.
     *
     * @param modelRspec Model that can be used as source for editing
     */
    public TimelineView(@Nonnull FXModelRspec modelRspec, @Nonnull NodePropertiesDialogFactory nodePropertiesDialog,
                        @Nonnull ImageUtil imageUtil, @Nonnull TimelineModelFactory timelineModelFactory, @Nonnull JFedGuiPreferences jFedPreferences) {
        this.modelRspec = modelRspec;
        controller = new TimelineController(jFedPreferences);
        model = timelineModelFactory.createControllerModel(modelRspec);
        controller.setModel(model);
        controller.setView(this);

        experimentController = new TimelineTimeline(model, controller);
        experimentList = new TimelineTreeView(model, controller, nodePropertiesDialog, imageUtil);

        experimentOutput = null;

        this.setDividerPosition(0, 0.2);
        this.setOrientation(Orientation.HORIZONTAL);
        this.getItems().addAll(experimentList, experimentController);

        SplitPane.setResizableWithParent(experimentList, false);

        initializeTimelineGui();

    }

    /**
     * Constructor for use as an "online" experiment recorder
     */
    public TimelineView(@Nonnull Experiment experiment, @Nonnull DetailsTabHost detailsTabHost,
                        @Nonnull NodePropertiesDialogFactory nodePropertiesDialog, @Nonnull ImageUtil imageUtil,
                        @Nonnull TimelineModelFactory timelineModelFactory, @Nonnull JFedGuiPreferences jFedPreferences) {
        FXModelRspec modelRspec = ExperimentUtils.getMergedRspec(experiment).getModelRspec(ModelRspecType.FX);
        if (modelRspec == null) {
            throw new RuntimeException("Cannot create TimelineView if RSpec is not valid");
        }
        this.modelRspec = modelRspec;
        controller = new TimelineController(jFedPreferences);
        model = timelineModelFactory.createControllerModel(experiment);
        controller.setModel(model);
        controller.setView(this);

        experimentController = new TimelineTimeline(model, controller);
        experimentList = new TimelineTreeView(model, controller, nodePropertiesDialog, imageUtil);

        experimentOutput = new ControllerOutput(detailsTabHost, controller, imageUtil);


        getItems().addAll(experimentList, experimentController);
        setOrientation(Orientation.HORIZONTAL);
        setDividerPosition(0, 0.2);


        SplitPane.setResizableWithParent(experimentList, false);

        initializeTimelineGui();
    }

    private void initializeTimelineGui() {
        if (modelRspec.getBarrierSegments().isEmpty()) {
            controller.addBarrierSegment();
        } else {
            for (ExperimentBarrierSegment b : modelRspec.getBarrierSegments()) {
                update(UpdateEvent.ADD_BARRIER, b);
            }
        }

        for (FXRspecNode n : model.getNodes()) {
            update(UpdateEvent.ADD_NODE, n);
            for (ExperimentCommand c : n.getCommands()) {
                LOG.debug("initializeTimelineGui() for command for node with id " + c.getNodeUniqueID() + " on node with id " + n.getUniqueId());
                assert Objects.equals(n.getUniqueId(), c.getNodeUniqueID()) :
                        "node id\"" + n.getUniqueId() + "\" does not equal command node id \"" + c.getNodeUniqueID() + "\"";
                update(UpdateEvent.ADD_COMMAND, c);
            }
        }
    }

    @Override
    public void update(UpdateEvent event, Object o) {
        experimentList.update(event, o);
        experimentController.update(event, o);

        //experimentOutput is not used in the editor mode
        if (experimentOutput != null && !model.isOfflineEditor()) {
            experimentOutput.update(event, o);
        }
    }

    @Nonnull
    public FXModelRspec getModelRspec() {
        return modelRspec;
    }

    @Nonnull
    public TimelineController getController() {
        return controller;
    }

    @Nonnull
    public TimelineModel getModel() {
        return model;
    }

    public ObservableBooleanValue onlineProperty() {
        return model.allConnectableProperty();
    }
}
